# -*- encoding:utf-8 -*-

from . import *


def sign(msg, secret0, k=None, canonical=True):
    """
    Generate signature according to ``ECDSA`` scheme.

    Args:
        msg (:class:`bytes`): sha256 message-hash
        secret0 (:class:`bytes`): private key
        k (:class:`int`): nonce (random nonce used if k=None)
        canonical (:class:`bool`): canonalize signature
    Returns:
        :class:`bytes`: DER signature
    """
    k = (rand_k() if not k else k) % n
    Q = G * k
    invk = pow(k, n-2, n)

    r = Q.x % n
    if r == 0:
        return None

    s = (invk * (int_from_bytes(msg) + int_from_bytes(secret0) * r)) % n
    if s == 0:
        return None
    if canonical and (s > (n//2)):
        s = n-s

    return der_from_sig(r, s)


def rfc6979_sign(msg, secret0, canonical=True):
    """
    Generate signature according to ``ECDSA`` scheme using a `RFC-6979 nonce <\
https://tools.ietf.org/html/rfc6979#section-3.2>`_

    Args:
        msg (:class:`bytes`): sha256 message-hash
        secret0 (:class:`bytes`): private key
        canonical (:class:`bool`): canonalize signature
    Returns:
        :class:`bytes`: DER signature
    """
    V = None
    for i in range(1, 10):
        k, V = rfc6979_k(msg, secret0, V)
        sig = sign(msg, secret0, k, canonical)
        if sig:
            return sig
    return None


def verify(msg, pubkey, sig):
    """
    Check signature according to ``ECDSA`` scheme.

    Args:
        msg (:class:`bytes`): sha256 message-hash
        pubkey (:class:`bytes`): encoded public key
        sig (:class:`bytes`): signature
    Returns:
        :class:`bool`: True if match
    """
    r, s = sig_from_der(sig)
    if r is None or r > n or s > n:
        return False

    h = int_from_bytes(msg)
    c = pow(s, n-2, n)

    u1G = G * ((h*c) % n)
    # u2Q = PublicKey.decode(pubkey) * ((r*c) % n)
    u2Q = point_mul(PublicKey.decode(pubkey), ((r*c) % n))
    GQ = u1G + u2Q

    return (GQ.x % n) == r



# -*- coding: utf-8 -*-
# © Toons

# import unittest
import pySecp256k1 as secp256k1
from pySecp256k1 import ecdsa
from pySecp256k1 import schnorr


class TestSecp256k1Signatures: #(unittest.TestCase):

    def testBcrypto410Schnorr(self):
        msg = secp256k1.hash_sha256("message to sign".encode())
        pr_key = secp256k1.hash_sha256("secret".encode())
        pu_key = secp256k1.encoded_from_point(
            secp256k1.G * secp256k1.int_from_bytes(pr_key)
        )
        assert schnorr.bcrypto410_verify(
                msg, pu_key,
                schnorr.bcrypto410_sign(msg, pr_key)
            ) is True

        # self.assertEqual(
        #     True,
        #     schnorr.bcrypto410_verify(
        #         msg, pu_key,
        #         schnorr.bcrypto410_sign(msg, pr_key)
        #     )
        # )

    def testSecp256k1Ecdsa(self, benchmark):
        msg = secp256k1.hash_sha256("message to sign".encode())
        pr_key = secp256k1.hash_sha256("secret".encode())
        pu_key = secp256k1.encoded_from_point(
            secp256k1.G * secp256k1.int_from_bytes(pr_key)
        )
        sig = benchmark(ecdsa.sign, msg, pr_key)
        # self.assertEqual(
        #     True,
        #     ecdsa.verify(msg, pu_key, sig)
        # )
        # self.assertEqual(
        #     True,
        #     ecdsa.verify(msg, pu_key, ecdsa.sign(msg, pr_key, canonical=False))
        # )

    def testSecp256k1Rfc6979Ecdsa(self):
        msg = secp256k1.hash_sha256("message to sign".encode())
        pr_key = secp256k1.hash_sha256("secret".encode())
        pu_key = secp256k1.encoded_from_point(
            secp256k1.G * secp256k1.int_from_bytes(pr_key)
        )
        # self.assertEqual(
        #     True,
        #     ecdsa.verify(msg, pu_key, ecdsa.rfc6979_sign(msg, pr_key))
        # )
        # self.assertEqual(
        #     True,
        #     ecdsa.verify(msg, pu_key, ecdsa.rfc6979_sign(
        #         msg, pr_key, canonical=False)
        #     )
        # )



# -*- encoding:utf-8 -*-

from . import *


# Note that bip schnorr uses a very different public key format (32 bytes) than
# the ones used by existing systems (which typically use elliptic curve points
# as public keys, 33-byte or 65-byte encodings of them). A side effect is that
# `PubKey(sk) = PubKey(bytes(n-int(sk))`, so every public key has two
# corresponding private keys.

def bytes_from_point(P):
    """
    Encode a public key as defined in `BIP schnorr <https://github.com/bitcoin\
/bips/blob/master/bip-0340.mediawiki>`_ spec.

    Args:
        P (:class:`Point`): secp256k1 curve point
    Returns:
        :class:`bytes`: encoded public key
    """
    return bytes_from_int(P[0])


def point_from_bytes(pubkeyB):
    """
    Decode a public key as defined in `BIP schnorr <https://github.com/bitcoin\
/bips/blob/master/bip-0340.mediawiki>`_ spec.

    Args:
        pubkeyB (:class:`bytes`): encoded public key
    Returns:
        :class:`Point`: secp256k1 curve point
    """
    x = int_from_bytes(pubkeyB)
    y = y_from_x(x)
    if not y:
        return None
    return [x, y]


# https://github.com/bcoin-org/bcrypto/blob/v4.1.0/lib/js/schnorr.js
def bcrypto410_sign(msg, seckey0):
    """
    Generate message signature according to `Bcrypto 4.10 schnorr <https://git\
hub.com/bcoin-org/bcrypto/blob/v4.1.0/lib/js/schnorr.js>`_ spec.

    Args:
        msg (:class:`bytes`): sha256 message-hash
        secret0 (:class:`bytes`): private key
    Returns:
        :class:`bytes`: RAW signature
    """
    # print(msg.hex())
    # print(seckey0.hex())
    if len(msg) != 32:
        raise ValueError('The message must be a 32-byte array.')

    seckey = int_from_bytes(seckey0)
    if not (1 <= seckey <= n - 1):
        raise ValueError(
            'The secret key must be an integer in the range 1..n-1.'
        )

    k0 = int_from_bytes(hash_sha256(seckey0 + msg)) % n
    if k0 == 0:
        raise RuntimeError(
            'Failure. This happens only with negligible probability.'
        )
    # print("[P]>>> hash = %s" % hash_sha256(seckey0 + msg).hex())
    # print("[P]>>> k = %x" % k0)

    R = G * k0
    Rraw = bytes_from_int(R.x)
    e = int_from_bytes(
        hash_sha256(Rraw + encoded_from_point(G*seckey) + msg)
    ) % n
    # print("[P]>>> e = %x" % e)

    # seckey %= n
    # k0 %= n
    k = n - k0 if not is_quad(R.y) else k0

    s = (k + e * seckey) % n
    # s %= n

    return Rraw + bytes_from_int(s)


def bcrypto410_verify(msg, pubkey, sig):
    """
    Check if public key match message signature according to `Bcrypto 4.10 sch\
norr <https://github.com/bcoin-org/bcrypto/blob/v4.1.0/lib/js/schnorr.js>`_
    spec.

    Args:
        msg (:class:`bytes`): sha256 message-hash
        pubkey (:class:`bytes`): encoded public key
        sig (:class:`bytes`): signature
    Returns:
        :class:`bool`: True if match
    """
    print(">>>", msg)
    if len(msg) != 32:
        raise ValueError('The message must be a 32-byte array.')
    if len(sig) != 64:
        raise ValueError('The signature must be a 64-byte array.')

    P = PublicKey.decode(pubkey)
    r, s = int_from_bytes(sig[:32]), int_from_bytes(sig[32:])
    if r >= p or s >= n:
        return False

    e = int_from_bytes(hash_sha256(sig[0:32] + pubkey + msg)) % n
    R = Point(*(G*s + point_mul(P, n-e)))  # P*(n-e) does not work...
    if R is None or not is_quad(R.y) or R.x != r:
        return False

    return True


# def sign(msg, seckey0):
#     """
#     Generate message signature according to `BIP schnorr <https://github.com/b\
# itcoin/bips/blob/master/bip-0340.mediawiki>`_ spec.

#     Args:
#         msg (:class:`bytes`): sha256 message-hash
#         seckey0 (:class:`bytes`): private key
#     Returns:
#         :class:`bytes`: RAW signature
#     """
#     if len(msg) != 32:
#         raise ValueError('The message must be a 32-byte array.')

#     seckey0 = int_from_bytes(seckey0)
#     if not (1 <= seckey0 <= n - 1):
#         raise ValueError(
#             'The secret key must be an integer in the range 1..n-1.'
#         )

#     P = G*seckey0
#     seckey = seckey0 if is_quad(P.y) else n - seckey0

#     k0 = int_from_bytes(
#         tagged_hash("BIPSchnorrDerive", bytes_from_int(seckey) + msg)
#     ) % n
#     if k0 == 0:
#         raise RuntimeError(
#             'Failure. This happens only with negligible probability.'
#         )

#     R = G*k0
#     k = n - k0 if not is_quad(R.y) else k0
#     r = bytes_from_point(R)
#     e = int_from_bytes(
#         tagged_hash("BIPSchnorr", r + bytes_from_point(P) + msg)
#     ) % n

#     return r + bytes_from_int((k + e * seckey) % n)


# def verify(msg, pubkey, sig):
#     """
#     Check if public key match message signature according to `BIP schnorr <htt\
# ps://github.com/bitcoin/bips/blob/master/bip-0340.mediawiki>`_ spec.

#     Args:
#         msg (:class:`bytes`): sha256 message-hash
#         pubkey (:class:`bytes`): encoded public key
#         sig (:class:`bytes`): signature
#     Returns:
#         :class:`bool`: True if match
#     """
#     if len(msg) != 32:
#         raise ValueError('The message must be a 32-byte array.')
#     if len(pubkey) != 32:
#         raise ValueError('The public key must be a 32-byte array.')
#     if len(sig) != 64:
#         raise ValueError('The signature must be a 64-byte array.')

#     P = point_from_bytes(pubkey)
#     if (P is None):
#         return False

#     r, s = int_from_bytes(sig[:32]), int_from_bytes(sig[32:])
#     if (r >= p or s >= n):
#         return False

#     e = int_from_bytes(tagged_hash("BIPSchnorr", sig[0:32] + pubkey + msg)) % n
#     R = Point(*(G*s + point_mul(P, n-e)))  # P*(n-e) does not work...
#     if R is None or not is_quad(R.y) or R.x != r:
#         return False

#     return True



# -*- encoding:utf-8 -*-

"""
Pure python implementation for `scp256k1` curve algebra and associated
`ECDSA - SCHNORR` signatures.

Sources:
 + [Schnorr](https://github.com/bitcoin/bips/blob/master/bip-0340.mediawiki)
 + [Reference](
     https://github.com/sipa/bips/blob/bip-taproot/bip-0340/reference.py
   )
 + [Bcrypto 4.10 schnorr](
     https://github.com/bcoin-org/bcrypto/blob/v4.1.0/lib/js/schnorr.js
   )

``python
>>> import pySecp256k1 as secp256k1
>>> secp256k1.G
<secp256k1 point:
  x:79be667ef9dcbbac55a06295ce870b07029bfcdb2dce28d959f2815b16f81798
  y:483ada7726a3c4655da4fbfc0e1108a8fd17b448a68554199c47d08ffb10d4b8
>
>>> G+G == 2*G
True
>>> secp256k1.PublicKey.from_int(
...     secp256k1.int_from_bytes(secp256k1.hash_sha256("secret"))
... )
<secp256k1 public key:
  x:a02b9d5fdd1307c2ee4652ba54d492d1fd11a7d1bb3f3a44c4a05e79f19de933
  y:924aa2580069952b0140d88de21c367ee4af7c4a906e1498f20ab8f62e4c2921
>
>>> secp256k1.PublicKey.from_seed(secp256k1.hash_sha256("secret"))
<secp256k1 public key:
  x:a02b9d5fdd1307c2ee4652ba54d492d1fd11a7d1bb3f3a44c4a05e79f19de933
  y:924aa2580069952b0140d88de21c367ee4af7c4a906e1498f20ab8f62e4c2921
>
>>> secp256k1.PublicKey.from_secret("secret")
<secp256k1 public key:
  x:a02b9d5fdd1307c2ee4652ba54d492d1fd11a7d1bb3f3a44c4a05e79f19de933
  y:924aa2580069952b0140d88de21c367ee4af7c4a906e1498f20ab8f62e4c2921
>
``

Variables:
  - `secret` (str): passphrase
  - `secret0` (bytes): private key
  - `P` (list): public key as `secp256k1` curve point
  - `pubkey` (bytes): compressed - encoded public key
  - `pubkeyB` (bytes): compressed - encoded public key according to bip
    schnorr spec
  - `msg` (bytes): sha256 hash of message to sign
  - Uppercase variables refer to points on the curve with equation `y²=x³+7`
    over the integers modulo `p`
"""

import hmac
import random
import hashlib
import future

from builtins import int, bytes, pow


p = int(0xfffffffffffffffffffffffffffffffffffffffffffffffffffffffefffffc2f)
n = int(0xfffffffffffffffffffffffffffffffebaaedce6af48a03bbfd25e8cd0364141)


def hash_sha256(b):
    """
    Arguments:
        b (bytes or str): sequence to be hashed
    Returns:
        sha256 hash bytes
    """
    return hashlib.sha256(
        b if isinstance(b, bytes) else b.encode("utf-8")
    ).digest()


#: precomputed hashtag
HASHED_TAGS = {
    "BIPSchnorrDerive": hash_sha256("BIPSchnorrDerive"),
    "BIPSchnorr": hash_sha256("BIPSchnorr"),
}


def _setNget(t):
    dict.__setitem__(HASHED_TAGS, t, hash_sha256(t))
    return dict.__getitem__(HASHED_TAGS, t)


def tagged_hash(tag, msg):
    """
    Return `sha256(sha256(tag) || sha256(tag) || msg)`. Tagged hash are
    registered to speed up code execution.

    Arguments:
        tag (str): tag to use
        msg (bytes): sha256 hash of message to sign
    Returns:
        tagged hash bytes
    """
    tag_hash = HASHED_TAGS.get(tag, _setNget(tag))
    return hash_sha256(tag_hash + tag_hash + msg)


def is_infinity(P):
    return P is None


def y_from_x(x):
    """
    Compute `y` from `x` according to `y²=x³+7`.
    """
    y_sq = (pow(x, 3, p) + 7) % p
    y = pow(y_sq, (p + 1) // 4, p)
    if pow(y, 2, p) != y_sq:
        return None
    return y


def point_on_curve(P):
    return y_from_x(P[0]) == P[1]


def point_add(P1, P2):
    """
    Add `secp256k1` points.

    Arguments:
        P1 (list): first `secp256k1` point
        P2 (list): second `secp256k1` point
    Returns:
        `secp256k1` point
    """
    if (P1 is None):
        return P2

    if (P2 is None):
        return P1

    xP1 = P1[0]
    yP1 = P1[1]
    xP2 = P2[0]
    yP2 = P2[1]

    if (xP1 == xP2):
        if yP1 != yP2:
            raise ValueError("One of the point is not on the curve")
        # P1 == P2
        else:
            lam = (3 * xP1 * xP1 * pow(2 * yP1, p - 2, p)) % p
    # P1 != P2
    else:
        lam = ((yP2 - yP1) * pow(xP2 - xP1, p - 2, p)) % p

    x3 = (lam * lam - xP1 - xP2) % p
    return [x3, (lam * (xP1 - x3) - yP1) % p]


def point_mul(P, n):
    """
    Multiply `secp256k1` point with scalar.

    Arguments:
        P (list): `secp256k1` point
        n (int): scalar
    Returns:
        `secp256k1` point
    """
    R = None
    for i in range(n.bit_length()):
        if ((n >> i) & 1):
            R = point_add(R, P)
        P = point_add(P, P)
    return R


def bytes_from_int(x):
    return int(x).to_bytes(32, byteorder="big")


def int_from_bytes(b):
    return int.from_bytes(b, byteorder="big")


def jacobi(x):
    return pow(x, (p - 1) // 2, p)


def is_quad(x):
    return jacobi(x) == 1


def has_square_y(P):
    return not is_infinity(P) and is_quad(P[1])


def encoded_from_point(P):
    """
    Encode and compress a `secp256k1` point:
      * `bytes(2) || bytes(x)` if y is even
      * `bytes(3) || bytes(x)` if y is odd

    Arguments:
        P (list): `secp256k1` point
    Returns:
        compressed and encoded point as bytes
    """
    return (b"\x03" if P[1] & 1 else b"\x02") + bytes_from_int(P[0])


def point_from_encoded(pubkey):
    """
    Decode and decompress a `secp256k1` point.

    Arguments:
        pubkey (bytes): compressed and encoded point
    Returns:
        `secp256k1` point
    """
    pubkey = bytearray(pubkey)
    x = int_from_bytes(pubkey[1:])
    y = y_from_x(x)
    if y is None:
        raise ValueError("Point not on `secp256k1` curve")
    elif y % 2 != pubkey[0] - 2:
        y = -y % p
    return [x, y]


def der_from_sig(r, s):
    """
    Encode a signature according `DER` spec.

    Arguments:
        r (int): signature part #1
        s (int): signature part #2
    Returns:
        encoded signature as bytes
    """
    r = bytes_from_int(r)
    s = bytes_from_int(s)
    r = (b'\x00' if (r[0] & 0x80) == 0x80 else b'') + r
    s = (b'\x00' if (s[0] & 0x80) == 0x80 else b'') + s
    return b'\x30' + int((len(r)+len(s)+4)).to_bytes(1, 'big') + \
           b'\x02' + int(len(r)).to_bytes(1, 'big') + r + \
           b'\x02' + int(len(s)).to_bytes(1, 'big') + s


def sig_from_der(der):
    """
    Decode a `DER` signature.

    Arguments:
        der (bytes): encoded signature
    Returns:
        signature (r, s) as int tuple
    """
    sig = bytearray(der)
    sig_len = sig[1] + 2
    r_offset, r_len = 4, sig[3]
    s_offset, s_len = 4+r_len+2, sig[4+r_len+1]
    if (
        sig[0] != 0x30 or sig_len != r_len+s_len+6 or sig[r_offset-2] != 0x02
        or sig[s_offset-2] != 0x02
    ):
        return None, None
    return (
        int_from_bytes(sig[r_offset:r_offset+r_len]),
        int_from_bytes(sig[s_offset:s_offset+s_len])
    )


def rand_k():
    """Generate a random nonce."""
    return random.getrandbits(p.bit_length()) % p


def rfc6979_k(msg, secret0, V=None):
    """
    Generate a deterministic nonce according to
    [rfc6979 spec](https://tools.ietf.org/html/rfc6979#section-3.2).

    Arguments:
        msg (bytes): 32-bytes sequence
        secret0 (bytes): private key
        V (bytes):
    Returns:
        deterministic nonce as int
    """
    hasher = hashlib.sha256
    if (V is None):
        # a.  Process m through the hash function H, yielding: h1 = H(m)
        h1 = msg
        hsize = len(h1)
        # b. Set: V = 0x01 0x01 0x01 ... 0x01
        V = b'\x01'*hsize
        # c. Set: K = 0x00 0x00 0x00 ... 0x00
        K = b'\x00'*hsize
        # d. Set: K = HMAC_K(V || 0x00 || int2octets(x) || bits2octets(h1))
        x = secret0
        K = hmac.new(K, V + b'\x00' + x + h1, hasher).digest()
        # e. Set: V = HMAC_K(V)
        V = hmac.new(K, V, hasher).digest()
        # f. Set: K = HMAC_K(V || 0x01 || int2octets(x) || bits2octets(h1))
        K = hmac.new(K, V + b'\x01' + x + h1, hasher).digest()
        # g. Set: V = HMAC_K(V)
        V = hmac.new(K, V, hasher).digest()

    # h.  Apply the following algorithm until a proper value is found for  k:
    while True:
        #
        # 1. Set T to the empty sequence.  The length of T (in bits) is
        #       denoted tlen; thus, at that point, tlen = 0.
        T = b''
        # 2. While tlen < qlen, do the following:
        #       V = HMAC_K(V)
        #       T = T || V
        p_blen = p.bit_length()
        while len(T)*8 < p_blen:
            V = hmac.new(K, V, hasher).digest()
            T = T + V
        # 3. Compute:
        k = int_from_bytes(T)
        k_blen = k.bit_length()

        if k_blen > p_blen:
            k = k >> (k_blen - p_blen)
        #      If that value of k is within the [1,q-1] range, and is
        #      suitable for DSA or ECDSA (i.e., it results in an r value
        #      that is not 0; see Section 3.4), then the generation of k is
        #      finished.  The obtained value of k is used in DSA or ECDSA.
        if k > 0 and k < (p-1):
            return k, V
        #      Otherwise, compute:
        #         K = HMAC_K(V || 0x00)
        #         V = HMAC_K(V)
        #         and loop (try to generate a new T, and so on).
        K = hmac.new(K, V+b'\x00', hasher).digest()
        V = hmac.new(K, V, hasher).digest()


class Point(list):
    """
    `secp256k1` point . Initialization can be done with sole `x` value.
    `secp256k1.Point` overrides `*` and `+` operators which accepts python
    list as argument and returns `secp256k1.Point`.
    """

    x = property(
        lambda cls: list.__getitem__(cls, 0),
        lambda cls, v: [
            list.__setitem__(cls, 0, int(v)),
            list.__setitem__(cls, 1, y_from_x(int(v)))
        ],
        None, "Return list item #0"
    )
    y = property(
        lambda cls: list.__getitem__(cls, 1),
        None, None, "Return list item #1"
    )

    def __init__(self, *xy):
        if len(xy) == 0:
            xy = (0, None)
        elif len(xy) == 1:
            xy += (y_from_x(int(xy[0])), )
        list.__init__(self, [int(e) if e is not None else e for e in xy[:2]])

    def __mul__(self, k):
        if isinstance(k, int):
            return Point(*point_mul(self, k))
        else:
            raise TypeError("'%s' should be an int" % k)
    __rmul__ = __mul__

    def __add__(self, P):
        if isinstance(P, list):
            return Point(*point_add(self, P))
        else:
            raise TypeError("'%s' should be a 2-int-length list" % P)
    __radd__ = __add__

    def __repr__(self):
        return "<secp256k1 point:\n  x:%064x\n  y:%064x\n>" % tuple(self)

    @staticmethod
    def decode(pubkey):
        """
        See :func:`point_from_encoded`.
        """
        return Point(*point_from_encoded(pubkey))

    def encode(self):
        """
        See :func:`encoded_from_point`.
        """
        return encoded_from_point(self)


class PublicKey(Point):
    """
    `secp256k1.Point` extension providing specific initialization methods.
    """

    @staticmethod
    def from_int(value):
        """
        Compute a public key from int value.

        Arguments:
            value (int): scalar to use
        Returns:
            the public key as `secp256k1.PublicKey`
        """
        if (1 <= value <= n - 1):
            return PublicKey(*(G * int(value)))
        else:
            raise ValueError(
                'The secret key must be an integer in the range 1..n-1.'
            )

    @staticmethod
    def from_seed(seed):
        """
        Compute a public key from bytes value.

        Arguments:
            value (bytes): bytes sequence to use
        Returns:
            the public key as `secp256k1.PublicKey`
        """
        return PublicKey.from_int(int_from_bytes(seed))

    @staticmethod
    def from_secret(secret):
        """
        Compute a public key from secret passphrase.

        Arguments:
            value (str): secret passphrase to use
        Returns:
            the public key as `secp256k1.PublicKey`
        """
        return PublicKey.from_seed(hash_sha256(secret))

    def __repr__(self):
        return "<secp256k1 public key:\n  x:%064x\n  y:%064x\n>" % tuple(self)


G = Point(0x79be667ef9dcbbac55a06295ce870b07029bfcdb2dce28d959f2815b16f81798)



# -*- coding: utf-8 -*-
# © Toons

import os
import io
import unittest
import binascii
from pySecp256k1 import schnorr


with io.open(
    os.path.join(os.path.dirname(__file__), "test_vectors.csv"), "r"
) as test_vectors:
    SCHNORR_TEST_VECTORS = test_vectors.read().split("\n")

VECTORS = []
header = [e.strip() for e in SCHNORR_TEST_VECTORS[0].split(";")]
for column in SCHNORR_TEST_VECTORS[1:]:
    VECTORS.append(dict(zip(header, column.split(";"))))


class TestSchnorrVectors(unittest.TestCase):

    def testVector1to4(self):
        for v in VECTORS[:4]:
            v = VECTORS[0]
            secret0 = binascii.unhexlify(v["secret key"])
            pubkey = binascii.unhexlify(v["public key"])
            msg = binascii.unhexlify(v["message"])
            sig = binascii.unhexlify(v["signature"])
            result = v["verification result"] == 'True'

            self.assertEqual(
                pubkey,
                schnorr.bytes_from_point(
                    schnorr.PublicKey.from_seed(secret0)
                )
            )
            self.assertEqual(sig, schnorr.sign(msg, secret0))
            self.assertEqual(result, schnorr.verify(msg, pubkey, sig))

    def testVector4(self):
        v = VECTORS[3]
        pubkey = binascii.unhexlify(v["public key"])
        msg = binascii.unhexlify(v["message"])
        sig = binascii.unhexlify(v["signature"])

        msg_mod_p = schnorr.bytes_from_int(
            schnorr.int_from_bytes(msg) % schnorr.p
        )
        self.assertEqual(False, schnorr.verify(msg_mod_p, pubkey, sig))
        msg_mod_n = schnorr.bytes_from_int(
            schnorr.int_from_bytes(msg) % schnorr.n
        )
        self.assertEqual(False, schnorr.verify(msg_mod_n, pubkey, sig))

    def testVector5(self):
        v = VECTORS[4]
        pubkey = binascii.unhexlify(v["public key"])
        msg = binascii.unhexlify(v["message"])
        sig = binascii.unhexlify(v["signature"])
        result = v["verification result"] == 'True'
        self.assertRaises(ValueError, schnorr.point_from_encoded, pubkey)
        self.assertEqual(result, schnorr.verify(msg, pubkey, sig))



# -*- coding:utf-8 -*-

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


with open("VERSION") as f1, open("README.md") as f2:
    VERSION = f1.read().strip()
    LONG_DESCRIPTION = f2.read()

kw = {
    "version": VERSION,
    "name": "pySecp256k1",
    "keywords": ["elliptic", "curve", "bitcoin"],
    "author": "Toons",
    "author_email": "moustikitos@gmail.com",
    "maintainer": "Toons",
    "maintainer_email": "moustikitos@gmail.com",
    "url": "https://github.com/Moustikitos/elliptic-curve",
    "download_url":
        "https://github.com/Moustikitos/elliptic-curve/archive/master.zip",
    "include_package_data": True,
    "description": "Pure python implementation for bitcoin curve",
    "long_description": LONG_DESCRIPTION,
    "long_description_content_type": "text/markdown",
    "packages": ["pySecp256k1"],
    "install_requires": ["future"],
    "tests_requires": ["pytest", "pytest-benchmark"],
    "license": "Copyright 2021, MIT licence",
    "classifiers": [
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
    ],
}

setup(**kw)



# -*- encoding:utf-8 -*-

"""
Advanced signature manipulation. It is the recomended module to manually issue
signatures.

Variables:
  - `privateKey` (str): hexlified private key
  - `publicKey` (str):  hexlified compressed - encoded public key
  - `message` (str):    message to sign as string
"""

import binascii
import pySecp256k1 as secp256k1
from pySecp256k1 import ecdsa, schnorr

_unhexlify = binascii.unhexlify


def puk(publicKey):
    if isinstance(publicKey, (secp256k1.Point, list, tuple)):
        return secp256k1.encoded_from_point(publicKey)
    elif isinstance(publicKey, str):
        return _unhexlify(publicKey)
    elif isinstance(publicKey, bytes):
        return publicKey
    else:
        raise ValueError("cannot encode publicKey")


class Signature(list):
    r = property(
        lambda cls: list.__getitem__(cls, 0),
        None, None, "Signature part #1"
    )

    s = property(
        lambda cls: list.__getitem__(cls, 1),
        None, None, "Signature part #2"
    )

    der = property(
        lambda cls: cls._der_getter(),
        lambda cls, v: setattr(cls, "_der", v),
        None, "Return DER encoded signature as bytes sequence"
    )

    raw = property(
        lambda cls: cls._raw_getter(),
        lambda cls, v: setattr(cls, "_raw", v),
        None, "Return RAW Encode signature as bytes sequence"
    )

    def __init__(self, *rs):
        list.__init__(self, [int(e) for e in rs])

    def __repr__(self):
        return "<secp256k1 signature:\n  r:%064x\n  s:%064x\n>" % tuple(self)

    def _der_getter(cls):
        if not hasattr(cls, "_der"):
            setattr(cls, "_der", secp256k1.der_from_sig(*cls))
        return getattr(cls, "_der")

    def _raw_getter(cls):
        if not hasattr(cls, "_raw"):
            setattr(
                cls, "_raw",
                secp256k1.bytes_from_int(cls[0]) +
                secp256k1.bytes_from_int(cls[1])
            )
        return getattr(cls, "_raw")

    def ecdsa_verify(self, message, publicKey):
        """
        Check if public key match message signature according to `ECDSA`
        scheme.

        Args:
            message (str): message to verify
            publicKey (str): public key
        Returns:
            True if match
        """
        return ecdsa.verify(
            secp256k1.hash_sha256(message),
            puk(publicKey),
            self.der
        )

    def b410_schnorr_verify(self, message, publicKey):
        """
        Check if public key match message signature according to
        [Bcrypto 4.10 schnorr](
            https://github.com/bcoin-org/bcrypto/blob/v4.1.0/lib/js/schnorr.js
        ) scheme.

        Args:
            message (str): message to verify
            publicKey (str): public key
        Returns:
            True if match
        """
        return schnorr.bcrypto410_verify(
            secp256k1.hash_sha256(message),
            puk(publicKey),
            self.raw
        )

    def schnorr_verify(self, message, publicKey):
        """
        Check if public key match message signature according to
        [BIP schnorr](
            https://github.com/sipa/bips/blob/bip-schnorr/bip-schnorr.mediawiki
        ) scheme.

        Args:
            message (str): message to verify
            publicKey (str): public key
        Returns:
            True if match
        """
        return schnorr.verify(
            secp256k1.hash_sha256(message),
            schnorr.bytes_from_point(
                secp256k1.Point.decode(puk(publicKey))
            ),
            self.raw
        )

    @staticmethod
    def from_raw(raw):
        """
        Decode signature from RAW encoded bytes sequence.

        Arguments:
            raw (bytes): encoded signature
        Returns:
            `secp256k1.sig.Signature`
        """
        length = len(raw) // 2
        sig = Signature(
            secp256k1.int_from_bytes(raw[:length]),
            secp256k1.int_from_bytes(raw[length:]),
        )
        sig._raw = raw
        return sig

    @staticmethod
    def from_der(der):
        """
        Decode signature from DER encoded bytes sequence.

        Arguments:
            der (bytes): encoded signature
        Returns:
            `secp256k1.sig.Signature`
        """
        sig = Signature(*secp256k1.sig_from_der(der))
        sig._der = der
        return sig

    @staticmethod
    def ecdsa_sign(message, privateKey, canonical=True):
        """
        Generate message signature according to `ECDSA` scheme using a random
        nonce.

        Args:
            message (str): message to verify
            privateKey (str): private key
            canonical (bool): canonalize signature
        Returns:
            `secp256k1.sig.Signature`
        """
        return Signature.from_der(
            ecdsa.sign(
                secp256k1.hash_sha256(message), _unhexlify(privateKey),
                canonical=canonical
            )
        )

    @staticmethod
    def ecdsa_rfc6979_sign(message, privateKey, canonical=True):
        """
        Generate message signature according to `ECDSA` scheme using a
        deterministic nonce (RFC-6976).

        Args:
            message (str): message to verify
            privateKey (str): private key
            canonical (bool): canonalize signature
        Returns:
            `secp256k1.sig.Signature`
        """
        return Signature.from_der(
            ecdsa.rfc6979_sign(
                secp256k1.hash_sha256(message), _unhexlify(privateKey),
                canonical=canonical
            )
        )

    @staticmethod
    def b410_schnorr_sign(message, privateKey):
        """
        Generate message signature according to
        [Bcrypto 4.10 schnorr](
            https://github.com/bcoin-org/bcrypto/blob/v4.1.0/lib/js/schnorr.js
        ) scheme.

        Args:
            message (str): message to verify
            privateKey (str): private key
        Returns:
            `secp256k1.sig.Signature`
        """
        return Signature.from_raw(
            schnorr.bcrypto410_sign(
                secp256k1.hash_sha256(message), _unhexlify(privateKey)
            )
        )

    @staticmethod
    def schnorr_sign(message, privateKey):
        """
        Generate message signature according to
        [BIP schnorr](
            https://github.com/sipa/bips/blob/bip-schnorr/bip-schnorr.mediawiki
        ) scheme.

        Args:
            message (str): message to verify
            privateKey (str): private key
        Returns:
            `secp256k1.sig.Signature`
        """
        return Signature.from_raw(
            schnorr.sign(
                secp256k1.hash_sha256(message), _unhexlify(privateKey)
            )
        )
